% Various utility functions that do not belong anywhere else.

\begin{code}
#include "gadts.h"
module Darcs.Progress ( beginTedious, endTedious, tediousSize,
                        debugMessage, withoutProgress,
                        progress, finishedOne, finishedOneIO,
                        progressList, progressFL, progressRL,
                        setProgressMode ) where

import Prelude hiding (lookup, catch)

import Control.Exception ( catch, throw )
import Control.Monad ( when )
import System.IO ( hFlush, stderr, hPutStr, hPutStrLn, hSetBuffering, BufferMode(LineBuffering) )
import System.IO.Unsafe ( unsafePerformIO )
import Data.Char ( toLower )
import Data.Map ( Map, empty, adjust, insert, delete, lookup )
import Data.Maybe ( isJust )
import Control.Concurrent ( ThreadId, forkIO, threadDelay )
import Data.IORef ( IORef, newIORef, readIORef, writeIORef, modifyIORef )

import Darcs.Patch.Ordered ( FL(..), RL(..), lengthRL, lengthFL )
import Darcs.Global ( verboseMode, debugMode )

debugMessage :: String -> IO ()
debugMessage m = when debugMode $ hPutStrLn stderr m

handleProgress :: String -> IO ()
handleProgress k = do threadDelay 1000000
                      mp <- getProgressData k
                      case mp of
                        Nothing -> return ()
                        Just _ -> handleMoreProgress k (-1)

handleMoreProgress :: String -> Int -> IO ()
handleMoreProgress k n = do mp <- getProgressData k
                            case mp of
                              Nothing -> return ()
                              Just p ->
                                  do when (n < sofar p) $ printProgress k p
                                     threadDelay 100000
                                     handleMoreProgress k (sofar p)

printProgress :: String -> ProgressData -> IO ()
printProgress k (ProgressData {sofar=s, total=Just t, latest=Just l}) =
    myput (k++" "++show s++"/"++show t++" : "++l) (k++" "++show s++"/"++show t)
printProgress k (ProgressData {latest=Just l}) =
    myput (k++" "++l) k
printProgress k (ProgressData {sofar=s, total=Just t}) =
    myput (k++" "++show s++"/"++show t) (k++" "++show s)
printProgress k (ProgressData {sofar=s}) =
    myput (k++" "++show s) k

myput :: String -> String -> IO ()
myput l s | '\n' `elem` l = if verboseMode then hPutStrLn stderr l
                                           else myput (takeWhile (/= '\n') l) s
myput l s = if length l > 80 || verboseMode then simpleput s
                                            else simpleput l

simpleput :: String -> IO ()
simpleput s = do hPutStr stderr $ '\r':s++"\r"
                 hFlush stderr
                 hPutStr stderr $ map (const ' ') s++"\r"
                 putStr $ '\r':map (const ' ') s++"\r"

beginTedious :: String -> IO ()
beginTedious k = do tid <- forkIO $ handleProgress k
                    debugMessage $ "Beginning " ++ lower k
                    setProgressData k $ ProgressData { sofar = 0,
                                                       latest = Nothing,
                                                       total = Nothing,
                                                       handler = Just tid }

endTedious :: String -> IO ()
endTedious k = when (progressMode) $ do p <- getProgressData k
                                        modifyIORef _progressData (delete k)
                                        when (isJust p) $ debugMessage $ "Done "++lower k

lower :: String -> String
lower (x:xs) = toLower x:xs
lower "" = ""

beginOrEndTedious :: String -> Int -> IO ()
beginOrEndTedious k l = do mp <- getProgressData k
                           case mp of
                             Nothing -> do beginTedious k
                                           tediousSize k l
                             Just p -> if total p == Just l
                                       then endTedious k
                                       else return ()

tediousSize :: String -> Int -> IO ()
tediousSize k s = updateProgressData k uptot
    where uptot p = case total p of Just t -> seq ts $ p { total = Just ts }
                                        where ts = t + s
                                    Nothing -> p { total = Just s }

minlist :: Int
minlist = 4

progressList :: String -> [a] -> [a]
progressList _ [] = []
progressList k (x:xs) = if l < minlist then x:xs
                                       else startit x : pl xs
    where l = length (x:xs)
          startit y = unsafePerformIO $ do beginOrEndTedious k l
                                           return y
          pl [y] = [startit y]
          pl [] = []
          pl (y:ys) = progress k y : pl ys

progressFL :: String -> FL a C(x y) -> FL a C(x y)
progressFL _ NilFL = NilFL
progressFL k (x:>:xs) = if l < minlist then x:>:xs
                                       else startit x :>: pl xs
    where l = lengthFL (x:>:xs)
          startit y = unsafePerformIO $ do beginOrEndTedious k l
                                           return y
          pl :: FL a C(x y) -> FL a C(x y)
          pl (y:>:NilFL) = (startit y) :>: NilFL
          pl NilFL = NilFL
          pl (y:>:ys) = progress k y :>: pl ys

progressRL :: String -> RL a C(x y) -> RL a C(x y)
progressRL _ NilRL = NilRL
progressRL k (x:<:xs) = if l < minlist then x:<:xs
                                       else startit x :<: pl xs
    where l = lengthRL (x:<:xs)
          startit y = unsafePerformIO $ do beginOrEndTedious k l
                                           return y
          pl :: RL a C(x y) -> RL a C(x y)
          pl (y:<:NilRL) = (startit y) :<: NilRL
          pl NilRL = NilRL
          pl (y:<:ys) = progress k y :<: pl ys

progress :: String -> a -> a
progress k a = unsafePerformIO $ progressIO k >> return a

progressIO :: String -> IO ()
progressIO k = updateProgressData k (\p -> p { sofar = sofar p + 1,
                                               latest = Nothing })

finishedOne :: String -> String -> a -> a
finishedOne k l a = unsafePerformIO $ finishedOneIO k l >> return a

finishedOneIO :: String -> String -> IO ()
finishedOneIO k l = updateProgressData k (\p -> p { sofar = sofar p + 1,
                                                    latest = Just l })

{-# NOINLINE _progressMode #-}
_progressMode :: IORef Bool
_progressMode = unsafePerformIO $ do hSetBuffering stderr LineBuffering
                                     newIORef True

{-# NOINLINE _progressData #-}
_progressData :: IORef (Map String ProgressData)
_progressData = unsafePerformIO $ newIORef empty

setProgressMode :: Bool -> IO ()
setProgressMode m = writeIORef _progressMode m

withoutProgress :: IO a -> IO a
withoutProgress j = do m <- readIORef _progressMode
                       setProgressMode False
                       a <- j `catch` \e -> setProgressMode m >> throw e
                       setProgressMode m
                       return a

updateProgressData :: String -> (ProgressData -> ProgressData) -> IO ()
updateProgressData k f = when (progressMode) $ modifyIORef _progressData (adjust f k)

setProgressData :: String -> ProgressData -> IO ()
setProgressData k p = when (progressMode) $ modifyIORef _progressData (insert k p)

getProgressData :: String -> IO (Maybe ProgressData)
getProgressData k = if progressMode then lookup k `fmap` readIORef _progressData
                                    else return Nothing

{-# NOINLINE progressMode #-}
progressMode :: Bool
progressMode = unsafePerformIO $ readIORef _progressMode

data ProgressData = ProgressData { sofar :: !Int,
                                   latest :: !(Maybe String),
                                   total :: !(Maybe Int),
                                   handler :: !(Maybe ThreadId)}

\end{code}
