%  Copyright (C) 2006-2007 David Roundy
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software Foundation,
%  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
\chapter{Hashed inventory format}
\label{hashed_format}

The hashed inventory format is similar to the ``DarcsRepo'' format (see
Chapter~\ref{repository_format}), but I haven't gotten around to
documenting it.

\begin{code}
{-# OPTIONS -fglasgow-exts #-}
#include "gadts.h"
module Darcs.Repository.HashedRepo ( revert_tentative_changes, finalize_tentative_changes,
                                     slurp_pristine, sync_repo,
                                     copy_pristine, copy_partials_pristine, pristine_from_working,
                                     apply_to_tentative_pristine, replacePristine,
                                     add_to_tentative_inventory, remove_from_tentative_inventory,
                                     read_repo, read_repo_lazily,
                                     write_tentative_inventory, copy_repo, slurp_all_but_darcs
                                   ) where

import System.Directory ( getDirectoryContents, removeFile )
import System.IO.Unsafe ( unsafeInterleaveIO )
import System.IO ( stderr, hPutStrLn )
import Data.List ( delete, (\\) )
import Control.Monad ( unless, when )

import Workaround ( renameFile, createDirectoryIfMissing )
import Darcs.Flags ( DarcsFlag( VerifyHash ) )
import Darcs.Patch.Set ( PatchSet )
import Darcs.FilePathUtils ( absolute_dir )
import Darcs.Repository.Prefs ( Cache, fetchFileUsingCache, writeFileUsingCache,
                                unionCaches, cleanCaches, repo2cache )
import Darcs.Repository.HashedIO ( applyHashed, slurpHashed, hashSlurped, listHashedContents,
                                   copyHashed, syncHashed, copyPartialsHashed )
import Darcs.Hopefully ( PatchInfoAnd, patchInfoAndPatch, n2pia, info,
                         extractHash, createHashed )
import Darcs.SlurpDirectory ( Slurpy, empty_slurpy, slurp_remove, slurp )
import Darcs.Patch ( RepoPatch, Patchy, Named, showPatch, patch2patchinfo, readPatch,
                     readPatchLazily )
import Darcs.Patch.Depends ( commute_to_end, slightly_optimize_patchset )
import Darcs.Patch.Info ( PatchInfo, showPatchInfo, human_friendly, readPatchInfo )
import FileName ( fp2fn )
import FastPackedString ( PackedString, nilPS, nullPS, readFilePS,
                          gzReadFilePS, takePS, dropPS,
                          tailPS, lengthPS,
                          packString, breakOnPS, unpackPS, dropWhitePS )
import Printer ( Doc, hcat, (<>), ($$), renderString, renderPS, text, invisiblePS )
import SHA1 ( sha1PS )
import Darcs.External ( copyFileOrUrl, cloneFile, fetchFilePS, Cachable( Uncachable ) )
import Darcs.Lock ( writeBinFile, writeDocBinFile, writeAtomicFilePS, appendBinFile, appendDocBinFile )
import Darcs.Utils ( withCurrentDirectory )
import Darcs.Progress ( beginTedious, endTedious, progressRL, debugMessage )
#include "impossible.h"
import Darcs.Patch.Ordered ( FL(..), RL(..), (:<)(..),
                             mapRL, mapFL, (+>+) )
import Darcs.Sealed ( Sealed, unsafeUnseal )

revert_tentative_changes :: IO ()
revert_tentative_changes =
    do cloneFile "_darcs/hashed_inventory" "_darcs/tentative_hashed_inventory"
       i <- gzReadFilePS "_darcs/hashed_inventory"
       writeBinFile "_darcs/tentative_pristine" $ "pristine:" ++ inv2pris i

finalize_tentative_changes :: RepoPatch p => Cache p -> [DarcsFlag] -> IO ()
finalize_tentative_changes c opts =
    do let t = "_darcs/tentative_hashed_inventory"
       -- first let's optimize it...
       read_tentative_repo c opts "." >>= write_tentative_inventory c
       -- then we'll add in the pristine cache,
       i <- gzReadFilePS t
       p <- gzReadFilePS "_darcs/tentative_pristine"
       writeDocBinFile t $ pris2inv (inv2pris p) i
       -- and rename it to its final value
       renameFile t "_darcs/hashed_inventory"
       -- then we'll remove obsolete bits of our pristine cache
       clean_pristine c opts "." "_darcs/hashed_inventory"
       -- and also clean out any global caches.
       cleanCaches c "pristine.hashed"

add_to_tentative_inventory :: RepoPatch p => Cache p -> PatchInfoAnd p -> IO FilePath
add_to_tentative_inventory c p =
    do hash <- snd `fmap` write_patch_if_necesary c p
       appendDocBinFile ("_darcs/tentative_hashed_inventory") $ showPatchInfo $ info p
       appendBinFile ("_darcs/tentative_hashed_inventory") $ "\nhash: " ++ hash ++ "\n"
       return $ "_darcs/patches/" ++ hash

remove_from_tentative_inventory :: RepoPatch p => Cache p -> [DarcsFlag]
                                -> FL (Named p) -> IO ()
remove_from_tentative_inventory cache opts to_remove =
       -- FIXME: This algorithm should be *far* simpler.  All we need do is
       -- to to remove the patches from a patchset and then write that
       -- patchset.  The commutation behavior of PatchInfoAnd should track
       -- which patches need to be rewritten for us.
    do allpatches <- read_tentative_repo cache opts "."
       let (_ :< skipped) = commute_to_end to_remove allpatches
       okay <- simple_remove_from_tentative_inventory cache opts
               (mapFL patch2patchinfo $ to_remove +>+ skipped)
       unless okay $ bug "bug in HashedRepo.remove_from_tentative_inventory"
       sequence_ $ mapFL (add_to_tentative_inventory cache . n2pia) skipped

simple_remove_from_tentative_inventory :: forall p. RepoPatch p =>
                                          Cache p -> [DarcsFlag] -> [PatchInfo] -> IO Bool
simple_remove_from_tentative_inventory cache opts pis = do
    inv <- read_tentative_repo cache opts "."
    case cut_inv pis (inv :: PatchSet p) of
      Nothing -> return False
      Just inv' -> do write_tentative_inventory cache inv'
                      return True
    where cut_inv [] x = Just x
          cut_inv x (NilRL:<:rs) = cut_inv x rs
          cut_inv xs ((hp:<:r):<:rs) | info hp `elem` xs = cut_inv (info hp `delete` xs) (r:<:rs)
          cut_inv _ _ = Nothing

writeHashFile :: Cache p -> String -> Doc -> IO String
writeHashFile c subdir d = writeFileUsingCache c subdir $ renderPS d

readHashFile :: Cache p -> [DarcsFlag] -> String -> String -> IO (String,PackedString)
readHashFile cache opts subdir hash =
    do (fn,hf) <- fetchFileUsingCache cache subdir hash
       when (VerifyHash `elem` opts && sha1PS hf /= hash) $
            fail $ "Hash failed hash:  "++ fn
       return (fn,hf)

read_repo :: RepoPatch p => Cache p -> [DarcsFlag] -> String -> IO (PatchSet p)
read_repo cache opts d = do
  realdir <- absolute_dir d
  read_repo_private cache opts False realdir "hashed_inventory" `catch`
              (\e -> do hPutStrLn stderr ("Invalid repository:  " ++ realdir)
                        ioError e)

read_tentative_repo :: RepoPatch p => Cache p -> [DarcsFlag] -> String -> IO (PatchSet p)
read_tentative_repo cache opts d = do
  realdir <- absolute_dir d
  read_repo_private cache opts False realdir "tentative_hashed_inventory" `catch`
              (\e -> do hPutStrLn stderr ("Invalid repository:  " ++ realdir)
                        ioError e)

read_repo_lazily :: RepoPatch p => Cache p -> [DarcsFlag] -> String -> IO (PatchSet p)
read_repo_lazily cache opts d = do
  realdir <- absolute_dir d
  read_repo_private cache opts True realdir "hashed_inventory" `catch`
              (\e -> do hPutStrLn stderr ("Invalid repository:  " ++ realdir)
                        ioError e)

read_repo_private :: RepoPatch p => Cache p -> [DarcsFlag] -> Bool
                  -> FilePath -> FilePath -> IO (PatchSet p)
read_repo_private cache opts am_lazy d iname =
 do inventories <- read_inventory_private cache opts (d++"/_darcs") iname

    let parse i h = do (fn,ps) <- readHashFile cache opts "patches" h
                       if am_lazy
                         then return $ fst $ readPatchLazily ps
                         else case readPatch ps of
                              Just (p,_) -> return p
                              Nothing -> fail $ unlines ["Couldn't parse file "++fn,
                                                         "which is patch",
                                                         renderString $ human_friendly i]
        parseinvs [] = return NilRL
        parseinvs (i:is) = do x <- read_patches parse i
                              xs <- unsafeInterleaveIO $ parseinvs is
                              return (progressRL ("Reading patches in "++d) x:<:xs)
    parseinvs inventories
    where read_patches :: RepoPatch p => (PatchInfo -> String -> IO (Sealed (Named p)))
                       -> [(PatchInfo,String)] -> IO (RL (PatchInfoAnd p))
          read_patches _ [] = return NilRL
          read_patches parse ((i,h):is) = do
              mp <- unsafeInterleaveIO $ createHashed h (fmap unsafeUnseal . parse i)
              rest <- read_patches parse is
              return $ (i `patchInfoAndPatch` mp) :<: rest

write_tentative_inventory :: RepoPatch p => Cache p -> PatchSet p -> IO ()
write_tentative_inventory c = write_either_inventory c "tentative_hashed_inventory"

copy_repo :: forall p. RepoPatch p => Cache p -> [DarcsFlag] -> String -> String -> IO ()
copy_repo cache opts inr outr = do
    createDirectoryIfMissing False (outr++"/_darcs/inventories")
    copyFileOrUrl opts (inr++"/_darcs/hashed_inventory") (outr++"/_darcs/hashed_inventory")
                  Uncachable -- no need to copy anything but hashed_inventory!
    appendBinFile (outr++"/_darcs/prefs/sources") (show $ repo2cache inr `unionCaches` cache)

write_either_inventory :: RepoPatch p => Cache p -> String -> PatchSet p -> IO ()
write_either_inventory c iname x =
    do createDirectoryIfMissing False "_darcs/inventories"
       hsh <- write_inventory_private c $ slightly_optimize_patchset x
       case hsh of
         Nothing -> writeBinFile ("_darcs/"++iname) ""
         Just h -> gzReadFilePS ("_darcs/inventories/"++h) >>= writeAtomicFilePS ("_darcs/"++iname)

write_inventory_private :: RepoPatch p => Cache p -> PatchSet p -> IO (Maybe String)
write_inventory_private _ NilRL = return Nothing
write_inventory_private _ (NilRL:<:NilRL) = return Nothing
write_inventory_private _ (NilRL:<:_) = -- This shouldn't be possible, so best to check...
    bug "malformed PatchSet in HashedRepo.write_inventory_private"
write_inventory_private c (x:<:xs) =
  do resthash <- write_inventory_private c xs
     inventory <- sequence $ mapRL (write_patch_if_necesary c) x
     let inventorylist = hcat (map pihash $ reverse inventory)
         inventorycontents = case resthash of
                             Just lasthash -> text ("Starting with inventory:\n"++lasthash) $$
                                              inventorylist
                             _ -> inventorylist
     hash <- writeHashFile c "inventories" inventorycontents
     return $ Just hash

write_patch_if_necesary :: RepoPatch p => Cache p -> PatchInfoAnd p C(x y) -> IO (PatchInfo, String)
write_patch_if_necesary c hp =
    case extractHash hp of
    Right h -> return (info hp, h)
    Left p -> fmap (\h -> (info hp, h)) $ writeHashFile c "patches" $ showPatch p

pihash :: (PatchInfo,String) -> Doc
pihash (pinf,hash) = showPatchInfo pinf $$ text ("hash: " ++ hash ++ "\n")

read_inventory_private :: Cache p -> [DarcsFlag] -> String -> String
                       -> IO [[(PatchInfo, String)]]
read_inventory_private cache opts d iname = do
    i <- skip_pristine `fmap` fetchFilePS (d++"/"++iname) Uncachable
    (rest,str) <- case breakOnPS '\n' i of
                  (swt,pistr) | swt == packString "Starting with inventory:" ->
                    case breakOnPS '\n' $ tailPS pistr of
                    (h,thisinv) | lengthPS h == 40 ->
                      do r <- unsafeInterleaveIO $ read_inventories
                              cache opts (unpackPS h)
                         return (r,thisinv)
                    _ -> fail $ "Bad hash in " ++ d ++ "/_darcs/" ++ iname
                  _ -> return ([],i)
    return $ reverse (read_patch_ids str) : rest

read_inventories :: Cache p -> [DarcsFlag] -> String -> IO [[(PatchInfo, String)]]
read_inventories cache opts ihash = do
    (fn,i_and_p) <- readHashFile cache opts "inventories" ihash
    let i = skip_pristine i_and_p
    (rest,str) <- case breakOnPS '\n' i of
                  (swt,pistr) | swt == packString "Starting with inventory:" ->
                    case breakOnPS '\n' $ tailPS pistr of
                    (h,thisinv) | lengthPS h == 40 ->
                      do r <- unsafeInterleaveIO $
                              read_inventories cache opts (unpackPS h)
                         return (r,thisinv)
                    _ -> fail $ "Bad hash in file " ++ fn
                  _ -> return ([],i)
    return $ reverse (read_patch_ids str) : rest

read_patch_ids :: PackedString -> [(PatchInfo, String)]
read_patch_ids inv | nullPS inv = []
read_patch_ids inv = case readPatchInfo inv of
                     Nothing -> []
                     Just (pinfo,r) ->
                         case readHash r of
                         Nothing -> []
                         Just (h,r') -> (pinfo,h) : read_patch_ids r'

readHash :: PackedString -> Maybe (String, PackedString)
readHash s = let s' = dropWhitePS s
                 (l,r) = breakOnPS '\n' s'
                 (kw,h) = breakOnPS ' ' l
             in if kw /= packString "hash:" || lengthPS h <= 1
                then Nothing
                else Just (unpackPS $ tailPS h,r)

{-
read_patch_ids :: (Stringalike s, ParserM m, Monad (m s)) => Bool -> m s [(PatchInfo, PackedString)]
read_patch_ids
          = do mn <- maybe_work readPatchInfo
               case mn of
                   Nothing -> return []
                   Just n ->
                       do d <- read_depends
                          Just p <- readPatch' want_eof
                          return $ NamedP n d p
-}

apply_pristine :: Patchy q => Cache p -> [DarcsFlag] -> String -> String -> q -> IO ()
apply_pristine c opts d iname p =
    do i <- gzReadFilePS (d++"/"++iname)
       h <- applyHashed c opts (inv2pris i) p
       writeDocBinFile (d++"/"++iname) $ pris2inv h i

apply_to_tentative_pristine :: Patchy q => Cache p -> [DarcsFlag] -> q -> IO ()
apply_to_tentative_pristine c opts p = apply_pristine c opts "." "_darcs/tentative_pristine" p

clean_pristine :: Cache p -> [DarcsFlag] -> String -> String -> IO ()
clean_pristine c opts d iname =
    withCurrentDirectory d $ do i <- gzReadFilePS iname
                                hs <- listHashedContents c opts $ inv2pris i
                                let hashdir = "_darcs/pristine.hashed/"
                                fs <- filter ((==40).length) `fmap` getDirectoryContents hashdir
                                mapM_ (removeFile . (hashdir++)) (fs \\ hs)

slurp_pristine :: Cache p -> [DarcsFlag] -> String -> String -> IO Slurpy
slurp_pristine c opts d iname = do
    i <- fetchFilePS (d++"/"++iname) Uncachable
    slurp_pristine_private c opts i

slurp_pristine_private :: Cache p -> [DarcsFlag] -> PackedString -> IO Slurpy
slurp_pristine_private c opts inv = case inv2pris inv of
                                    h | h == sha1PS nilPS -> return empty_slurpy
                                      | otherwise -> slurpHashed c opts h

pristine_from_working :: Cache p -> [DarcsFlag] -> IO ()
pristine_from_working c opts = replacePristine c opts "."

replacePristine :: Cache p -> [DarcsFlag] -> FilePath -> IO ()
replacePristine c opts d = do s <- slurp_all_but_darcs d
                              h <- hashSlurped c opts s
                              let t = "_darcs/hashed_inventory"
                              i <- gzReadFilePS t
                              writeDocBinFile t $ pris2inv h i

copy_pristine :: Cache p -> [DarcsFlag] -> String -> String -> IO ()
copy_pristine c opts d iname = do
    i <- fetchFilePS (d++"/"++iname) Uncachable
    debugMessage $ "Copying hashed pristine tree: "++inv2pris i
    let k = "Copying pristine"
    beginTedious k
    copyHashed k c opts $ inv2pris i
    endTedious k

sync_repo :: Cache p -> IO ()
sync_repo c = do i <- readFilePS "_darcs/hashed_inventory"
                 s <- slurp_all_but_darcs "."
                 syncHashed c s $ inv2pris i
                 

copy_partials_pristine :: Cache p -> [DarcsFlag] -> String -> String -> [FilePath] -> IO ()
copy_partials_pristine c opts d iname fps =
  do i <- fetchFilePS (d++"/"++iname) Uncachable
     copyPartialsHashed c opts (inv2pris i) fps

inv2pris :: PackedString -> String
inv2pris inv | takePS pristine_name_length inv == pristine_name =
                 unpackPS $ takePS 40 $ dropPS pristine_name_length inv
             | otherwise = sha1PS nilPS

pris2inv :: String -> PackedString -> Doc
pris2inv h inv = invisiblePS pristine_name <> text h $$ invisiblePS (skip_pristine inv)

pristine_name :: PackedString
pristine_name = packString "pristine:"

skip_pristine :: PackedString -> PackedString
skip_pristine ps
    | takePS pristine_name_length ps == pristine_name = dropPS (pristine_name_length+41) ps
    | otherwise = ps

pristine_name_length :: Int
pristine_name_length = lengthPS pristine_name

slurp_all_but_darcs :: FilePath -> IO Slurpy
slurp_all_but_darcs d = do s <- slurp d
                           case slurp_remove (fp2fn "./_darcs") s of
                             Nothing -> return s
                             Just s' -> return s'

\end{code}
