#include <stdio.h>
#include <string.h>
#include <WWWLib.h>
#include <WWWInit.h>

enum RESULT_CODES
  {
    RESULT_OK = 0,
    // Start from 1000 to avoid conflict with libwww codes.
    // libwww codes are defined in HTUtils.h
    RESULT_REQUEST_NEW_FAIL = 1000,
    RESULT_REQUEST_ADD_AFTER_FAIL,
    RESULT_REQUEST_ADD_CACHE_CONTROL_FAIL,
    RESULT_LOAD_TO_CHUNK_FAIL,
    RESULT_FILE_OPEN_FAIL,
    RESULT_FILE_WRITE_FAIL,
    RESULT_MALLOC_FAIL,
    RESULT_LIST_NEW_FAIL,
    RESULT_LIST_APPEND_OBJECT_FAIL
  };

struct Completed
{
  int error;
  char *url;
};

static BOOL init_done = NO;
static const char *const user_agent = "darcs";
static int error;
static HTList *completed_list = NULL;
static char *last_url = NULL;

int terminate_handler(HTRequest *request,
                      HTResponse *response,
                      void *param,
                      int status)
{
  HTChunk *const chunk = HTRequest_context(request);

  struct Completed *completed = malloc(sizeof(struct Completed));
  if (completed == NULL)
    error = RESULT_MALLOC_FAIL;
  else
    {
      if (HTList_appendObject(completed_list, completed) == NO)
        {
          error = RESULT_LIST_APPEND_OBJECT_FAIL;
          free(completed);
        }
      else
        {
          error = RESULT_OK;
          if (status == HT_LOADED)
            {
              completed->error = RESULT_OK;
              FILE *const f = fopen(param, "wb");
              if (f != NULL)
                {
                  if (HTChunk_size(chunk) > 0)
                    {
                      if (fwrite(HTChunk_data(chunk),
                                 HTChunk_size(chunk), 1, f) != 1)
                        completed->error = RESULT_FILE_WRITE_FAIL;
                    }
                  fclose(f);
                }
              else
                completed->error = RESULT_FILE_OPEN_FAIL;
            }
          else
            completed->error = status;

          completed->url = HTAnchor_address((HTAnchor *)HTRequest_anchor(request));
        }
    }

  HTChunk_delete(chunk);
  HTRequest_delete(request);
  free(param);

  HTEventList_stopLoop();

  // Return not HT_OK to stop executing after filters.
  return HT_ALL;
}

int libwww_request_url(const char *darcs_version,
                       const char *proxypass,
                       const char *url,
                       const char *filename,
                       int cache_time)
{
  error = RESULT_OK;
  BOOL result;

  if (init_done == NO)
    {
      HTProfile_newNoCacheClient(user_agent, darcs_version);
      HTProxy_getEnvVar();
      HTAlert_setInteractive(NO);
      HTFormat_addCoding("*", NULL, HTIdentityCoding, 1.0);
#if 0
      HTSetTraceMessageMask("sop");
#endif
      init_done = YES;
    }

  if (completed_list == NULL)
    {
      completed_list = HTList_new();
      if (completed_list == NULL)
        return error = RESULT_LIST_NEW_FAIL;
    }

  HTRequest *const request = HTRequest_new();
  if (request == NULL)
    error = RESULT_REQUEST_NEW_FAIL;

  result = HTRequest_addAfter(request, terminate_handler, NULL, strdup(filename),
                              HT_ALL, HT_FILTER_LAST, YES);
  if (result == NO)
    return error = RESULT_REQUEST_ADD_AFTER_FAIL;

  HTRequest_setOutputFormat(request, WWW_SOURCE);
  if (cache_time == 0)
    {
      HTRequest_addGnHd(request, HT_G_PRAGMA_NO_CACHE);
      result = HTRequest_addCacheControl(request, "no-cache", "");
    }
  else if (cache_time > 0)
    {
      char buf[8];
      snprintf(buf, sizeof(buf), "%d", cache_time);
      buf[sizeof(buf) - 1] = '\0';
      result = HTRequest_addCacheControl(request, "max-age", buf);
    }
  if (result == FALSE)
    return error = RESULT_REQUEST_ADD_CACHE_CONTROL_FAIL;

  HTChunk *const chunk = HTLoadToChunk(url, request);
  if (chunk == NULL)
    return error = RESULT_LOAD_TO_CHUNK_FAIL;
  else
    HTRequest_setContext(request, chunk);

  return error;
}

int libwww_wait_next_url()
{
  if (last_url != NULL)
    {
      free(last_url);
      last_url = NULL;
    }

  error = RESULT_OK;
  if (HTList_isEmpty(completed_list) == YES &&
      HTNet_isEmpty() == NO){
    HTEventList_newLoop();
  }

  if (HTList_isEmpty(completed_list) == NO)
    {
      struct Completed *completed = HTList_firstObject(completed_list);
      error = completed->error;
      last_url = completed->url;
      HTList_removeFirstObject(completed_list);
      free(completed);
    }

  return error;
}

char *libwww_last_url()
{
  return last_url != NULL ? last_url : "";
}
